/*****************************************************************
***        (C)Copyright 2015, NetApp, Inc.                     ***
***                                                            ***
***                    All Rights Reserved                     ***
***                                                            ***
******************************************************************
* Support FPGA Transactional FIFO Driver
* This driver code will only support LML
******************************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/cacheflush.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ide.h>
#include <crc32.h>

#include "Types.h"
#include "agentlite.h"
#include "support_fpga.h"
#include "support_fpga_hw.h"
#include "support_fpga_defs.h"

#include <sfpga_tfifo.h>

#define SFPGA_TFIFO_MAJOR_NUM  84

/* The size of the local message fifo */
#define LOCAL_FIFO_SIZE 4096

#define tfifo_msg_overflow(pt)  ((pt->status) & TFIFO_STAT_OVERFLOW)
#define tfifo_msg_present(pt)   ((pt->local_fifo.read_index != pt->local_fifo.write_index) || \
                                ((pt->status) & TFIFO_STAT_CMTD_MSK))
#define tfifo_committed_wr(pt)  ((pt->status) & TFIFO_STAT_CMTD_MSK)

static void tfifo_debug(const char *, ...) __attribute__((format(printf, 1, 2)));

static char hsam_sp_fault_buf[HSAM_MAX_MSG_SIZE];

/* Local FIFO structure */
typedef struct {
	int  write_index;  /* Index of next byte to write */
	int  read_index;   /* Index of next byte to read */
	char fifo_data[LOCAL_FIFO_SIZE];
} local_fifo_t;

/* Transactional FIFO Driver Private data */
typedef struct {
	struct semaphore    read_mutex;
	struct semaphore    ioctl_mutex;
	wait_queue_head_t   zz;  // Wait queue
	uint16_t            (*read_reg)(uint32_t reg_offset);
	void                (*write_reg)(uint32_t reg_offset, uint16_t sfpga_tfifo_value);
	local_fifo_t        local_fifo;
	tfifo_msg_hdr_t    *sp_fault_msg;
	uint32_t            delimiter;
	uint16_t            status;
} sfpga_tfifo_drv_t;

static inline uint16_t
tfifo_get_status(sfpga_tfifo_drv_t *p) {
	return p->read_reg(SFPGA_TFIFO_REG_STS);
}
static inline void
tfifo_clr_overflow(sfpga_tfifo_drv_t *p) {
	p->write_reg(SFPGA_TFIFO_REG_STS, TFIFO_STAT_OVERFLOW);
}
static inline uint16_t
tfifo_get_data(sfpga_tfifo_drv_t *p) {
	return p->read_reg(SFPGA_TFIFO_REG_DATA);
}
static inline bool
tfifo_irq_trig(sfpga_tfifo_drv_t *p) {
	return (p->read_reg(SFPGA_IRQ_STS_REG(3)) & TFIFO_IRQ_BIT);
}
static inline void
tfifo_reset(sfpga_tfifo_drv_t *p) {
	p->write_reg(SFPGA_TFIFO_REG_DATA, TFIFO_CTRL_RESET);
}
static inline void
tfifo_read_start(sfpga_tfifo_drv_t *p) {
	p->write_reg(SFPGA_TFIFO_REG_DATA, TFIFO_CTRL_RD_START);
}
static inline void
tfifo_read_commit(sfpga_tfifo_drv_t *p) {
	p->write_reg(SFPGA_TFIFO_REG_DATA, TFIFO_CTRL_RD_END);
}
static inline void
tfifo_irq_enable(sfpga_tfifo_drv_t *p) {
	p->write_reg(SFPGA_IRQ_EN_REG(3), p->read_reg(SFPGA_IRQ_EN_REG(3)) | TFIFO_IRQ_ENABLE);
}
static inline void
tfifo_irq_disable(sfpga_tfifo_drv_t *p) {
	p->write_reg(SFPGA_IRQ_EN_REG(3), p->read_reg(SFPGA_IRQ_EN_REG(3)) & ~TFIFO_IRQ_ENABLE);
}

static UINT32
calculate_crc32(unsigned char *buf, UINT32 Size)
{
	UINT32 i, crc32 = ~0;

	/* Read the data and calculate crc32 */
	for(i = 0; i < Size; i++)
		crc32 = ((crc32) >> 8) ^ CrcLookUpTable[(buf[i]) ^ ((crc32) & 0x000000FF)];
	return ~crc32;
}

/* -------------------------------------------------------------------------
 *  Static debug variables
 * ------------------------------------------------------------------------- */
#define DEBUG_BUF_LEN 512
#define DEBUG_ENTRY_MAX (DEBUG_BUF_LEN - 128)
static char debug_buf[DEBUG_BUF_LEN];
static int dbIndex = 0;
static struct mutex tfifo_debug_lock;

static uint16_t tfifo_last_data_reg;
static tfifo_msg_hdr_t db_tmsg;

/* -------------------------------------------------------------------------
 *  Static variables
 * ------------------------------------------------------------------------- */
static struct file_operations sfpga_tfifo_ops;

static uint32_t sfpga_tfifo_read_count = 0;
static uint32_t sfpga_tfifo_irq_handled_cnt = 0;
static uint32_t sfpga_tfifo_ioctl_count = 0;
static uint32_t sfpga_tfifo_open_count = 0;
static uint32_t sfpga_tfifo_close_count = 0;
static uint32_t crc32_msg = 0;
static uint32_t crc32_hdr = 0;

struct tasklet_struct tfifo_tasklet;
struct proc_dir_entry *sfpga_tfifo_proc = NULL;

sfpga_tfifo_drv_t *tfifo_pdata = NULL;

static char crc_buf[HSAM_MAX_MSG_SIZE];
static char msg_buf[HSAM_MAX_MSG_SIZE];

/* read and write 32 byte */
static inline void
sp_reg_write(volatile uint32_t *addr, uint32_t val)
{
	*addr = val;
}

static inline volatile uint32_t
sp_reg_read(volatile uint32_t *addr)
{
	return *addr;
}

static void
tfifo_debug(const char *fmt, ...)
{
	int i, j;
	va_list alst;

	mutex_lock(&tfifo_debug_lock);

	while (dbIndex > DEBUG_ENTRY_MAX) {
		for (i = 0; i < DEBUG_BUF_LEN; i++) {
			if (debug_buf[i] == '\n') {
				i++;
				break;
			}
		}
		if (i >= dbIndex) {
			dbIndex = 0;
		} else {
			for (j = 0; i < dbIndex; i++, j++) {
				debug_buf[j] = debug_buf[i];
			}
			dbIndex = j;
		}
	}

	va_start(alst, fmt);
	dbIndex += snprintf(debug_buf + dbIndex, DEBUG_BUF_LEN - dbIndex, fmt, alst);
	va_end(alst);

	mutex_unlock(&tfifo_debug_lock);

	return;
}

static uint16_t
sfpga_tfifo_read_reg(uint32_t offset) {
	uint32_t val = 0;

	support_fpga_read_parallel_bus_reg(&val, offset);
	return (uint16_t)val;
}

static void
sfpga_tfifo_write_reg(uint32_t offset, uint16_t value) {
	support_fpga_write_parallel_bus_reg((uint32_t)value, offset);
	return;
}

static void
tfifo_task(unsigned long privdata)
{
	sfpga_tfifo_drv_t *ptfifo = (sfpga_tfifo_drv_t *)privdata;
	/* Wake up the sleeping process */
	wake_up_interruptible(&ptfifo->zz);
	return;
}

static irqreturn_t
tfifo_data_pending_isr(int irq, void *priv)
{
	volatile uint32_t ior = 0;
	volatile uint32_t isr  = 0;
	sfpga_tfifo_drv_t *ptfifo = (sfpga_tfifo_drv_t *)priv;
	irqreturn_t tfifo_irq = IRQ_NONE;

	ior = sp_reg_read(INT_OBSERVE_REG);
	/* Check if the interrupt is generated by GPIO interface */
	if(ior & GPIO_OBSERVE_INT) {
		/* Check if the interrupt  was caused by GPIO P4_5 */
		isr = sp_reg_read(TFIFO_GPIO_IRQ_STS);
		if ((isr & TFIFO_IRQ_PIN) && tfifo_irq_trig(ptfifo))
		{
			tfifo_irq_disable(ptfifo);
			tfifo_irq = IRQ_HANDLED;
			sfpga_tfifo_irq_handled_cnt++;
			sp_reg_write(TFIFO_GPIO_IRQ_STS, TFIFO_IRQ_PIN);
			tasklet_init(&tfifo_tasklet, tfifo_task, (unsigned long)priv);
			tasklet_schedule(&tfifo_tasklet);
		}
	}

	return tfifo_irq;
}

static int
sfpga_tfifo_open(struct inode *inode, struct file *file) {
	sfpga_tfifo_open_count++;

	if (tfifo_pdata) {
		file->private_data = tfifo_pdata;
	}
	return 0;
}

static int
sfpga_tfifo_close(struct inode *inode, struct file *file) {
	sfpga_tfifo_close_count++;

	if (file->private_data) {
		file->private_data = NULL;
	}

	return 0;
}

/* Read a single byte from TFIFO */
static int
tfifo_read_byte(sfpga_tfifo_drv_t *ptfifo, uint16_t *p)
{
	uint16_t dreg;

	dreg = ptfifo->read_reg(SFPGA_TFIFO_REG_DATA);
	tfifo_last_data_reg = dreg;
	if (tfifo_data_ok(dreg)) {
		*p = tfifo_extract_data(dreg);
		return 0;
	}
	tfifo_debug("TFIFO data is invalid\n");
	return -ENOMSG;
}

/* Local FIFO is a byte-stream used in the same manner as TFIFO, except that
 * all messages are sent by the local processor. As these messages are stored
 * in memory, no byte-swapping is reqiuired
 *
 * Stack variables:
 *  ptfifo      Driver private data (sfpga_tfifo_drv_t)
 *  rcv_buf     Pointer to buffer to hold received data
 *  rcv_size    Size of the data received
 *  buf_size    Size of the receive buffer
 *
 * Returned variable:
 *  rc          zero == passed, non-zero == errorcode from reading local FIFO
 */
static int
lfifo_used_bytes(int read_index, int write_index) {
	/* Calculate free space in the fifo:
	 *   w: write index
	 *   r: read index
	 *   _: unused bytes
	 *   #: used bytes
	 *
	 *   w == r  ->  used = 0
	 *   w >  r  ->  used = write - read           ex: |___r#####w__|
	 *   w <  r  ->  used = max  + write - read    ex: |###w_____r##|
	 */
	int used = write_index - read_index;
	if(used < 0) {
		used += LOCAL_FIFO_SIZE;
	}
	return used;
}


/* Write data into the local fifo, do not change the send and receive pointers
 * Returns 0 on success, non-zero on error
 */
static int
lfifo_write_bytes(sfpga_tfifo_drv_t *ptfifo, int *write_index, char *buffer, int size) {
	int rc = -ENOSPC;
	int free_size;
	int chunk_size;
	char *fifo_data = ptfifo->local_fifo.fifo_data;
	int read_index = ptfifo->local_fifo.read_index;

	/* Calculate free space in the fifo */
	free_size = LOCAL_FIFO_SIZE - lfifo_used_bytes(read_index, *write_index);

	if(size < free_size) {
		/* We have enough space to do the write */
		chunk_size = LOCAL_FIFO_SIZE - *write_index;
		if(chunk_size > size) {
			/* We have plenty of space, can copy in one chunk */
			chunk_size = size;
		}
		memcpy(&fifo_data[*write_index], buffer, chunk_size);
		memcpy(&fifo_data[0], &buffer[chunk_size], size - chunk_size);
		*write_index = (*write_index + size) % LOCAL_FIFO_SIZE;
		rc = 0;
	}
	return rc;
}

static int
lfifo_read_bytes(sfpga_tfifo_drv_t *ptfifo, int *read_index, char *buffer, int size) {
	int rc = -ENODATA;
	int used_size;
	int chunk_size;
	char *fifo_data = ptfifo->local_fifo.fifo_data;
	int write_index = ptfifo->local_fifo.write_index;

	/* Calculate used space in the fifo */
	used_size = lfifo_used_bytes(*read_index, write_index);

	if(size <= used_size) {
		/* We have enough data to do the read */
		chunk_size = LOCAL_FIFO_SIZE - *read_index;
		if(chunk_size > size) {
			/* We have plenty of space, can copy in one chunk */
			chunk_size = size;
		}
		memcpy(buffer, &fifo_data[*read_index], chunk_size);
		memcpy(&buffer[chunk_size], &fifo_data[0], size - chunk_size);
		*read_index = (*read_index + size) % LOCAL_FIFO_SIZE;
		rc = 0;
	}
	return rc;
}

static int
lfifo_read_message(sfpga_tfifo_drv_t *ptfifo, void *buffer, int buffer_size)
{
	int rc = -ENODATA;
	int read_index = ptfifo->local_fifo.read_index;
	int size;

	/* Read size of this message */
	if(lfifo_read_bytes(ptfifo, &read_index, (char*)&size, sizeof(size)) == 0) {
		if(size > buffer_size) {
			/* Buffer not large enough to contain message */
			rc = -ENOSPC;
		} else {
			/* Read message body */
			if(lfifo_read_bytes(ptfifo, &read_index, buffer, size) == 0) {
				/* message read successfully */
				ptfifo->local_fifo.read_index = read_index;
				rc = 0;
			}
		}
	}
	return rc;
}

static int
lfifo_write_message(sfpga_tfifo_drv_t *ptfifo, void *buffer, int size)
{
	int rc = -ENOSPC;
	int write_index = ptfifo->local_fifo.write_index;

	if(/* Write size of this message */
	   (lfifo_write_bytes(ptfifo, &write_index, (char*)&size, sizeof(size)) == 0) &&
	   /* Write message body */
	   (lfifo_write_bytes(ptfifo, &write_index, buffer, size) == 0))
	{
		/* Write completed sucessfully */
		ptfifo->local_fifo.write_index = write_index;
		rc = 0;
	}
	return rc;
}

/* TFIFO is a byte-stream and header data needs to be byte-swapped.
 * This function read each header field from TFIFO and store it in 2 buffers
 * passed in as variables. One buffer would be byte-swapped and one buffer
 * would retain the original byte-stream for checksum purpose.
 *
 * Stack variables:
 *  ptfifo      Driver private data (sfpga_tfifo_drv_t)
 *  buf         Pointer to header field in CPU native (swapped for big endian)
 *  crc_buf     Pointer to header field in byte-stream (original data)
 *  sz          Size of the header field to be read
 *
 * Returned variable:
 *  rc          zero == passed, non-zero == errorcode from reading TFIFO
 */
static int
tfifo_read_header_field(sfpga_tfifo_drv_t *ptfifo, void *buf, char *crc_buf, int sz)
{
	uint16_t datareg;
	char *p;
	int i =0, rc = 0;

	p = (char *)buf;
	while (sz) {
		rc = tfifo_read_byte(ptfifo, &datareg);

		if (rc) {
			break;
		}
		sz--;
		p[sz] = (char)datareg;
		crc_buf[i++] = (char)datareg;
	}

	return rc;
}

static int
tfifo_read_data(sfpga_tfifo_drv_t *ptfifo, void *buf, int sz)
{
	uint16_t datareg;
	char *p;
	int i;
	int rc = 0;

	p = (char *)buf;
	for (i = 0; i < sz; i++) {
		rc = tfifo_read_byte(ptfifo, &datareg);
		if (rc) {
			break;
		}
		p[i] = (char)datareg;
	}

	return rc;
}

static ssize_t
tfifo_get_fault_message(sfpga_tfifo_drv_t *ptfifo, char *buf, int buf_size)
{
	tfifo_msg_hdr_t *tmsg;
	char            *msg_body;
	uint32_t        crc_temp;
	int             rc=0;
	int             i;

	tmsg = (tfifo_msg_hdr_t *)buf;

	/* Start a TFIFO read command */
	tfifo_read_start(ptfifo);
	i = 0;

	if((buf == NULL) || (buf_size < sizeof(tmsg->delimiter))) {
		tfifo_debug("TFIFO invalid arguments\n");
		goto tfifo_get_fault_message_done;
	}
	/* Read HSAM message header from transactional FIFO */
	db_tmsg.delimiter = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->delimiter, &crc_buf[i], sizeof(tmsg->delimiter));
	i += sizeof(tmsg->delimiter);
	if (rc) {
		tfifo_debug("TFIFO read delimiter failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.delimiter = tmsg->delimiter;

	tmsg->msglen = 0;
	db_tmsg.msglen = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->msglen, &crc_buf[i], sizeof(tmsg->msglen));
	i += sizeof(tmsg->msglen);
	if (rc) {
		tfifo_debug("TFIFO read message len failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.msglen = tmsg->msglen;

	db_tmsg.msgid = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->msgid, &crc_buf[i], sizeof(tmsg->msgid));
	i += sizeof(tmsg->msgid);
	if (rc) {
		tfifo_debug("TFIFO read message ID failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.msgid = tmsg->msgid;

	db_tmsg.origin = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->origin, &crc_buf[i], sizeof(tmsg->origin));
	i += sizeof(tmsg->origin);
	if (rc) {
		tfifo_debug("TFIFO read originated ID failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.origin = tmsg->origin;

	tmsg->hdrlen = 0;
	db_tmsg.hdrlen = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->hdrlen, &crc_buf[i], sizeof(tmsg->hdrlen));
	i += sizeof(tmsg->hdrlen);
	if (rc) {
		tfifo_debug("TFIFO read header len failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.hdrlen = tmsg->hdrlen;

	db_tmsg.version = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->version, &crc_buf[i], sizeof(tmsg->version));
	i += sizeof(tmsg->version);
	if (rc) {
		tfifo_debug("TFIFO read version failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.version = tmsg->version;

	db_tmsg.reserve = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->reserve, &crc_buf[i], sizeof(tmsg->reserve));
	i += sizeof(tmsg->reserve);
	if (rc) {
		tfifo_debug("TFIFO read reserve failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.reserve = tmsg->reserve;

	db_tmsg.crc32hdr = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->crc32hdr, &crc_buf[i], sizeof(tmsg->crc32hdr));
	i += sizeof(tmsg->crc32hdr);
	if (rc) {
		tfifo_debug("TFIFO read header crc failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.crc32hdr = tmsg->crc32hdr;

	db_tmsg.crc32msg = 0;
	rc = tfifo_read_header_field(ptfifo, &tmsg->crc32msg, &crc_buf[i], sizeof(tmsg->crc32msg));
	i += sizeof(tmsg->crc32msg);
	if (rc) {
		tfifo_debug("TFIFO read message crc failed\n");
		goto tfifo_get_fault_message_done;
	}
	db_tmsg.crc32msg = tmsg->crc32msg;

	if (!tfifo_msg_valid(ptfifo, tmsg->delimiter)) {
		tfifo_debug("TFIFO invalid delimiter\n");
		rc = -ENOMSG;
		goto tfifo_get_fault_message_done;
	}
	if (tmsg->msglen < sizeof(tfifo_msg_hdr_t) || tmsg->msglen > buf_size) {
		tfifo_debug("TFIFO invalid message len\n");
		rc = -ENOMSG;
		goto tfifo_get_fault_message_done;
	}
	if (tmsg->hdrlen < sizeof(tfifo_msg_hdr_t) || tmsg->hdrlen > tmsg->msglen) {
		tfifo_debug("TFIFO invalid header len\n");
		rc = -ENOMSG;
		goto tfifo_get_fault_message_done;
	} else if (tmsg->hdrlen != sizeof(tfifo_msg_hdr_t)) {
		/* To support backward compatibility, we need to read in the
		 * the rest of unknown header fields to pass checksum */
		rc = tfifo_read_header_field(ptfifo, &buf[sizeof(tfifo_msg_hdr_t)], &crc_buf[i], tmsg->hdrlen - sizeof(tfifo_msg_hdr_t));
		i += tmsg->hdrlen - sizeof(tfifo_msg_hdr_t);
	}

	/* Read HSAM message body from transactional FIFO */
	msg_body = buf + tmsg->hdrlen;
	rc = tfifo_read_data(ptfifo, msg_body, tmsg->msglen - tmsg->hdrlen);
	if (rc) {
		tfifo_debug("TFIFO read message body failed\n");
		goto tfifo_get_fault_message_done;
	}
	memcpy(&crc_buf[i], msg_body, tmsg->msglen - tmsg->hdrlen);

	crc_temp = ((tfifo_msg_hdr_t *)crc_buf)->crc32hdr;
	((tfifo_msg_hdr_t *)crc_buf)->crc32hdr = 0;
	((tfifo_msg_hdr_t *)crc_buf)->crc32msg = 0;
	crc32_hdr = calculate_crc32(crc_buf, tmsg->hdrlen);
	((tfifo_msg_hdr_t *)crc_buf)->crc32hdr = crc_temp;

	crc32_msg = calculate_crc32(crc_buf, tmsg->msglen);

	if (crc32_hdr != tmsg->crc32hdr) {
		tfifo_debug("TFIFO header checksum mismatched\n");
		rc = -ENOMSG;
		goto tfifo_get_fault_message_done;
	}
	if (crc32_msg != tmsg->crc32msg) {
		tfifo_debug("TFIFO message checksum mismatched\n");
		rc = -ENOMSG;
		goto tfifo_get_fault_message_done;
	}

tfifo_get_fault_message_done:
	if (!rc) {
		rc = tmsg->msglen;
	} else {
		/* If failed, reset uncommitted count to leave data in TFIFO.
		 * User can use ioctl to search, flush or reset TFIFO */
		tfifo_read_start(ptfifo);
	}
	tfifo_read_commit(ptfifo);
	return rc;
}

static ssize_t
sfpga_tfifo_read(struct file *pfile, char *user_buf, size_t buf_size, loff_t *poffset)
{
	wait_queue_t        read_wait;
	ssize_t             rc = 0;
	char               *kernel_buf = NULL;
	static char temp_buf[4096];
	sfpga_tfifo_drv_t  *ptfifo = pfile->private_data;

	if (down_interruptible(&ptfifo->read_mutex)) {
		tfifo_debug("TFIFO read: cannot acquire read mutex\n");
		return -ERESTARTSYS;
	}

	/* Enable GPIO interrupt for SFPGA line #3 */
	sp_reg_write(TFIFO_GPIO_IRQ_ADR, INPUT_INTR_FALL_EDGE);

	/* Blocking ioctl since we don't want a race with ioctl */
	if (down_interruptible(&ptfifo->ioctl_mutex)) {
		tfifo_debug("TFIFO read: cannot acquire ioctl mutex\n");
		rc = -ERESTARTSYS;
		goto sfpga_tfifo_read_error;
	}

	sfpga_tfifo_read_count++;

	ptfifo->status = tfifo_get_status(ptfifo);

	if (tfifo_msg_overflow(ptfifo)) {
		/* Only display here for debugging.
		 * User application should clear this via ioctl command */
		tfifo_debug("TFIFO read: overflow detected\n");
	}

	if (!tfifo_msg_present(ptfifo)) {
		/* Return if the process wants Non-blocking I/O */
		if(pfile->f_flags & O_NONBLOCK) {
			tfifo_debug("TFIFO read: no data available on non-blocking call\n");
			rc = -EAGAIN;
			goto sfpga_tfifo_read_exit;
		}

		/* Blocking I/O */
		init_waitqueue_entry(&read_wait, current);
		add_wait_queue(&ptfifo->zz, &read_wait);
		while (!tfifo_msg_present(ptfifo)) {
			set_current_state(TASK_INTERRUPTIBLE);
			ptfifo->status = tfifo_get_status(ptfifo);
			if (!tfifo_msg_present(ptfifo)) {
				tfifo_debug("TFIFO read: waiting for message\n");
				up(&ptfifo->ioctl_mutex);
				tfifo_irq_enable(ptfifo);
				schedule();
				tfifo_irq_disable(ptfifo);
				if (down_interruptible(&ptfifo->ioctl_mutex)) {
					tfifo_debug("TFIFO read: cannot acquire ioctl mutex\n");
					remove_wait_queue(&ptfifo->zz, &read_wait);
					rc = -ERESTARTSYS;
					goto sfpga_tfifo_read_error;
				}
			} else {
				break;
			}

			if (signal_pending(current)) {
				tfifo_debug("TFIFO read: signal interrupted\n");
				rc = -EINTR;
				remove_wait_queue(&ptfifo->zz, &read_wait);
				goto sfpga_tfifo_read_exit;
			}
		}
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&ptfifo->zz, &read_wait);
	}

	if (buf_size < sizeof(tfifo_msg_hdr_t)) {
		tfifo_debug("TFIFO read: message buf %d is too small\n", buf_size);
		rc = -EINVAL;
		goto sfpga_tfifo_read_exit;
	}
	if (buf_size > HSAM_MAX_MSG_SIZE) {
		buf_size = HSAM_MAX_MSG_SIZE;
	}
	kernel_buf = temp_buf;
	if (!kernel_buf) {
		tfifo_debug("TFIFO read: cannot allocate kernel space\n");
		rc = -ENOMEM;
		goto sfpga_tfifo_read_exit;
	}

	/* If we have a message from the local side, process that first */
	if (lfifo_read_message(ptfifo, kernel_buf, buf_size) == 0) {
		/* Save message attributes for debug output */
		db_tmsg.delimiter = ((tfifo_msg_hdr_t*)kernel_buf)->delimiter;
		db_tmsg.msglen    = ((tfifo_msg_hdr_t*)kernel_buf)->msglen;
		db_tmsg.msgid     = ((tfifo_msg_hdr_t*)kernel_buf)->msgid;
		db_tmsg.origin    = ((tfifo_msg_hdr_t*)kernel_buf)->origin;
		db_tmsg.hdrlen    = ((tfifo_msg_hdr_t*)kernel_buf)->hdrlen;
		db_tmsg.version   = ((tfifo_msg_hdr_t*)kernel_buf)->version;
		db_tmsg.reserve   = 0;
		db_tmsg.crc32hdr  = 0;
		db_tmsg.crc32msg  = 0;
		rc = buf_size;
	}
	/* Otherwise, process message from transactional fifo */
	else {
		rc = tfifo_get_fault_message(ptfifo, kernel_buf, buf_size);
	}
	if (rc > 0) {
		if (copy_to_user(user_buf, kernel_buf, rc)) {
			tfifo_debug("TFIFO read: copy to user failed %s(%d), user: 0x%px kernel: 0x%px size: %d\n",
				__FUNCTION__, __LINE__, user_buf, kernel_buf, rc);
			rc = -EFAULT;
		}
	}

sfpga_tfifo_read_exit:
	up(&ptfifo->ioctl_mutex);

sfpga_tfifo_read_error:
	/* Disable GPIO interrupt */
	sp_reg_write(TFIFO_GPIO_IRQ_ADR, INPUT_INTR_DISABLE);
	up(&ptfifo->read_mutex);

	return(rc);
}

static long
sfpga_tfifo_ioctl( struct file *file, uint cmd, ulong arg)
{
	tfifo_msg_hdr_t tmsg;
	tfifo_ioctl_msg_t reg;
	tfifo_ioctl_msg_t msg;
	sfpga_tfifo_drv_t *ptfifo;
	char *kernel_buf;
	uint32_t len;
	uint32_t tdata;
	uint16_t tmp;
	int i, rc = 0;

	sfpga_tfifo_ioctl_count++;

	ptfifo = file->private_data;

	if (!ptfifo->read_reg || !ptfifo->write_reg) {
		tfifo_debug("TFIFO ioctl: no register read/write implementation %s(%d)\n", __FUNCTION__, __LINE__);
		return -ENOSYS;
	}

	if (down_interruptible(&ptfifo->ioctl_mutex)) {
		tfifo_debug("TFIFO ioctl: cannot acquire ioctl mutex\n");
		return -ERESTARTSYS;
	}

	ptfifo->status = tfifo_get_status(ptfifo);
	len = tfifo_committed_wr(ptfifo);

	switch (cmd) {
	case e_TFIFO_REG_READ:
		copy_from_user((void *)&reg, (void *)arg, sizeof(tfifo_ioctl_msg_t));
		if (tfifo_valid_reg(reg.offset)) {
			reg.offset += SFPGA_TFIFO_OFFSET;
			reg.value = ptfifo->read_reg(reg.offset);
			copy_to_user((void *)arg, (void *)&reg, sizeof(tfifo_ioctl_msg_t));
		} else {
			tfifo_debug("TFIFO ioctl: read offset 0x%x out of range\n", reg.offset);
			rc = -EINVAL;
		}
		break;

	case e_TFIFO_REG_WRITE:
		copy_from_user((void *)&reg, (void *)arg, sizeof(tfifo_ioctl_msg_t));
		if (tfifo_valid_reg(reg.offset)) {
			reg.offset += SFPGA_TFIFO_OFFSET;
			ptfifo->write_reg(reg.offset, reg.value);
		} else {
			tfifo_debug("TFIFO ioctl: write offset 0x%x out of range\n", reg.offset);
			rc = -EINVAL;
		}
		break;

	case e_TFIFO_BUFSIZE:
		reg.offset = 0;
		reg.value = SFPGA_TFIFO_SIZE;
		copy_to_user((void *)arg, (void *)&reg, sizeof(tfifo_ioctl_msg_t));
		break;

	case e_TFIFO_MSG_INSERT:
		copy_from_user((void *)&msg, (void *)arg, sizeof(tfifo_ioctl_msg_t));
		if (msg.length < sizeof(tfifo_msg_hdr_t) || msg.length > HSAM_MAX_MSG_SIZE) {
			tfifo_debug("TFIFO ioctl: command message length %d out of range\n", msg.length);
			rc = -ENOSPC;
			break;
		}
		copy_from_user((void *)&tmsg, (void *)msg.message, sizeof(tfifo_msg_hdr_t));
		if (tmsg.msglen < sizeof(tfifo_msg_hdr_t) || tmsg.msglen > HSAM_MAX_MSG_SIZE) {
			tfifo_debug("TFIFO ioctl: message length %d out of range\n", tmsg.msglen);
			rc = -ENOSPC;
			break;
		}
		if (tmsg.hdrlen < sizeof(tfifo_msg_hdr_t) || tmsg.hdrlen > tmsg.msglen) {
			tfifo_debug("TFIFO ioctl: header length %d out of range\n", tmsg.hdrlen);
			rc = -EINVAL;
			break;
		}
		copy_from_user(msg_buf, (void *)msg.message, tmsg.msglen);
		rc = lfifo_write_message(ptfifo, msg_buf, tmsg.msglen);
		if(rc != 0) {
			tfifo_debug("TFIFO ioctl: FIFO full, cannot insert message\n");
		}
		/* Wake up the sleeping process */
		wake_up_interruptible(&ptfifo->zz);
		break;

	case e_TFIFO_OVERFLOW_CLR:
		tfifo_clr_overflow(ptfifo);
		break;

	case e_TFIFO_DELIMITER_SRCH:
		if (len > SFPGA_TFIFO_SIZE) {
			tfifo_debug("TFIFO ioctl: HW error, size 0x%x is too big\n", len);
			rc = -EIO;
			break;
		}
		/* How many delimiter occurrences in the search? */
		copy_from_user((void *)&reg, (void *)arg, sizeof(tfifo_ioctl_msg_t));
		i = reg.offset > 0? reg.offset : 1;	/* minimum is 1 */

		/* If the search is success, value hold delimiter and offset
		 * hold index into TFIFO of first encountered delimiter */
		reg.offset = 0;
		reg.value = 0;
		tfifo_read_start(ptfifo);
		while (reg.offset < len && i) {
			rc = tfifo_read_byte(ptfifo, &tmp);
			tdata = (uint32_t)tmp;
			if (rc) {
				break;
			}
			reg.value = (reg.value >> 8) | (tdata << 24);
			reg.offset++;
			if (reg.value == ptfifo->delimiter) {
				i--;
			}
		}
		if (i == 0) {
			reg.offset -= sizeof(ptfifo->delimiter);
		}

		/* We only want to search for delimiter without extracting any
		 * data. TFIFO-read-restart would reset uncommitted count and
		 * leave data in TFIFO. */
		tfifo_read_start(ptfifo);
		tfifo_read_commit(ptfifo);
		copy_to_user((void *)arg, (void *)&reg, sizeof(tfifo_ioctl_msg_t));
		break;

	case e_TFIFO_DELIMITER_SET:
		copy_from_user((void *)&reg, (void *)arg, sizeof(tfifo_ioctl_msg_t));
		/* delimiter must not be 0 or -1 */
		if (reg.value && ~reg.value) {
			tfifo_pdata->delimiter = reg.value;
		} else {
			tfifo_debug("TFIFO ioctl: invalid delimiter 0x%x\n", reg.value);
			rc = -EINVAL;
		}
		break;

	case e_TFIFO_FLUSH:
		if (len > SFPGA_TFIFO_SIZE) {
			tfifo_debug("TFIFO ioctl: HW error, size 0x%x is too big\n", len);
			rc = -EIO;
			break;
		}
		copy_from_user((void *)&msg, (void *)arg, sizeof(tfifo_ioctl_msg_t));
		/* Only flush what user requested or
		 * flush the entire FIFO if no instruction from user */
		if (msg.length && msg.length < len) {
			len = msg.length;
		}
		kernel_buf = (char *)kmalloc(len, GFP_KERNEL);
		if (!kernel_buf) {
			tfifo_debug("TFIFO ioctl cannot allocate kernel space\n");
			rc = -ENOMEM;
			break;
		}
		tfifo_read_start(ptfifo);
		rc = tfifo_read_data(ptfifo, kernel_buf, len);
		tfifo_read_commit(ptfifo);
		copy_to_user((void *)arg, (void *)&len, sizeof(len));
		copy_to_user((void *)msg.message, (void *)kernel_buf, len);
		kfree(kernel_buf);
		break;

	case e_TFIFO_RESET:
		tfifo_reset(ptfifo);
		break;

	default:
		printk (KERN_ALERT "Invalid TFIFO driver ioctl command\n");
		rc = -EINVAL;
		break;
	}

	up(&ptfifo->ioctl_mutex);

	return rc;
}

static int
sfpga_tfifo_get_info(char *sysbuf, char **p_sysbuf, off_t offset, int l_sysbuf, int *eof, void *unused)
{
	static int len = 0;

	if(offset != 0) {
		/* When offset != 0, Linux is calling us for more data,
		 * which we don't have any more so return 0 */
		len = 0;
		return 0;
	}

	/* When offset == 0, Linux is calling us for new data.
	 * Build the single big string in the kernel's buffer. */
	len += snprintf(sysbuf+len, l_sysbuf - len, "*** %s(%d) ***\n", SFPGA_TFIFO_DRIVER_NAME, SFPGA_TFIFO_MAJOR_NUM);
	len += snprintf(sysbuf+len, l_sysbuf - len, "opens:   %10d\n", sfpga_tfifo_open_count);
	len += snprintf(sysbuf+len, l_sysbuf - len, "closes:  %10d\n", sfpga_tfifo_close_count);
	len += snprintf(sysbuf+len, l_sysbuf - len, "reads:   %10d\n", sfpga_tfifo_read_count);
	len += snprintf(sysbuf+len, l_sysbuf - len, "ioctls:  %10d\n", sfpga_tfifo_ioctl_count);
	len += snprintf(sysbuf+len, l_sysbuf - len, "tfifo_irqs: %7d\n", sfpga_tfifo_irq_handled_cnt);
	len += snprintf(sysbuf+len, l_sysbuf - len, "\nTFIFO Last Message...\n");
	len += snprintf(sysbuf+len, l_sysbuf - len, "delimiter: 0x%08x\n", db_tmsg.delimiter);
	len += snprintf(sysbuf+len, l_sysbuf - len, "msglen:    %10d\n", db_tmsg.msglen);
	len += snprintf(sysbuf+len, l_sysbuf - len, "msgid:         0x%04x\n", db_tmsg.msgid);
	len += snprintf(sysbuf+len, l_sysbuf - len, "origin:          0x%02x\n", db_tmsg.origin);
	len += snprintf(sysbuf+len, l_sysbuf - len, "hdrlen:    %10d\n", db_tmsg.hdrlen);
	len += snprintf(sysbuf+len, l_sysbuf - len, "version:   %10d\n", db_tmsg.version);
	len += snprintf(sysbuf+len, l_sysbuf - len, "reserve:         0x%02x\n", db_tmsg.reserve);
	len += snprintf(sysbuf+len, l_sysbuf - len, "crc32hdr:  0x%08x\n", db_tmsg.crc32hdr);
	len += snprintf(sysbuf+len, l_sysbuf - len, "crc32msg:  0x%08x\n", db_tmsg.crc32msg);
	len += snprintf(sysbuf+len, l_sysbuf - len, "cmpt hdr:  0x%08x\n", crc32_hdr);
	len += snprintf(sysbuf+len, l_sysbuf - len, "cmpt msg:  0x%08x\n", crc32_msg);
	len += snprintf(sysbuf+len, l_sysbuf - len, "Last data rd:  0x%04x\n", tfifo_last_data_reg);

	mutex_lock(&tfifo_debug_lock);
	len += snprintf(sysbuf+len, l_sysbuf - len, "\n%s\n", debug_buf);
	dbIndex = 0;
	debug_buf[0] = 0;
	mutex_unlock(&tfifo_debug_lock);

	*eof = 1;

	if (len >= l_sysbuf) {
		/* The default kernel buffer is 1024 so we should never get here */
		printk (KERN_ALERT "sysbuf overflowed, len: %d, sysbuf: %d\n", len, l_sysbuf);
	}

	return len;
}

static void
sfpga_tfifo_exit(void) {
	printk (KERN_ALERT "Transactional FIFO driver exit\n");

	if (tfifo_pdata) {
		free_irq(TFIFO_IRQ_GPIO, tfifo_pdata);
	}

	unregister_chrdev(SFPGA_TFIFO_MAJOR_NUM, SFPGA_TFIFO_DRIVER_NAME);

	if (tfifo_pdata) {
		kfree(tfifo_pdata);
		tfifo_pdata = NULL;
	}

	remove_proc_entry(SFPGA_TFIFO_DRIVER_NAME, NULL);
	sfpga_tfifo_proc = NULL;

	return;
}

static int
sfpga_tfifo_init(void)
{
	int err, sts = 0;

	/* Create a proc handler */
	sfpga_tfifo_proc = create_proc_entry(SFPGA_TFIFO_DRIVER_NAME, S_IFREG | S_IRUGO, NULL);
	if (!sfpga_tfifo_proc) {
		printk (KERN_ERR "TFIFO Init: failed to create " SFPGA_TFIFO_DRIVER_NAME " proc entry\n");
		return -ENXIO;
	}
	sfpga_tfifo_proc->nlink = 1;
	sfpga_tfifo_proc->read_proc = sfpga_tfifo_get_info;

	mutex_init(&tfifo_debug_lock);

	memset(&sfpga_tfifo_ops,0,sizeof(sfpga_tfifo_ops));
	sfpga_tfifo_ops.open = sfpga_tfifo_open;
	sfpga_tfifo_ops.release = sfpga_tfifo_close;
	sfpga_tfifo_ops.unlocked_ioctl = sfpga_tfifo_ioctl;
	sfpga_tfifo_ops.read = sfpga_tfifo_read;
	sfpga_tfifo_ops.write = NULL;
	sfpga_tfifo_ops.owner = THIS_MODULE;

	/* Register the character device */
	if ((err = register_chrdev(SFPGA_TFIFO_MAJOR_NUM, SFPGA_TFIFO_DRIVER_NAME, &sfpga_tfifo_ops)) < 0) {
		printk (KERN_ERR "TFIFO Init: register_chrdrv failed. err = %d\n", err);
		remove_proc_entry(SFPGA_TFIFO_DRIVER_NAME, NULL);
		return -ENODEV;
	}

	tfifo_pdata = (sfpga_tfifo_drv_t*)kmalloc(sizeof(sfpga_tfifo_drv_t), GFP_KERNEL);
	if (!tfifo_pdata) {
		printk (KERN_ERR "TFIFO Init: cannot allocate kernel space\n");
		sts = -ENOMEM;
		goto sfpga_tfifo_init_done;
	}
	memset(tfifo_pdata, 0, sizeof(sfpga_tfifo_drv_t));

	/* Initialize the private data */
	tfifo_pdata->read_reg = sfpga_tfifo_read_reg;
	tfifo_pdata->write_reg = sfpga_tfifo_write_reg;
	sema_init(&tfifo_pdata->read_mutex, 1);
	sema_init(&tfifo_pdata->ioctl_mutex, 1);
	init_waitqueue_head(&tfifo_pdata->zz);
	tfifo_irq_disable(tfifo_pdata);
	tfifo_pdata->sp_fault_msg = (tfifo_msg_hdr_t *)hsam_sp_fault_buf;
	tfifo_pdata->delimiter = TFIFO_MSG_DELIMITER;
	tfifo_pdata->status = 0;

	sts = request_irq(TFIFO_IRQ_GPIO, tfifo_data_pending_isr, IRQF_SHARED, SFPGA_TFIFO_DRIVER_NAME, tfifo_pdata);
	if (sts) {
		printk(KERN_ERR "Failed to request irq %d (err %d)\n", TFIFO_IRQ_GPIO, sts);
	}

sfpga_tfifo_init_done:
	if (sts != 0) {
		sfpga_tfifo_exit();
	}

	return sts;
}

module_init(sfpga_tfifo_init);
module_exit(sfpga_tfifo_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Transactional FIFO Driver");
